/* (c) 1996-2003, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "backtalk.h"

#include <ctype.h>

#include "format.h"
#include "dict.h"
#include "sysdict.h"
#include "stack.h"
#include "exec.h"
#include "str.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

/* vstring is private to format.c - defines a reallocable string */
typedef struct vstring {
    char *begin;
    char *end;
    size_t size;	/* end - begin */
    char *p;		/* current pointer */
    char *last;		/* last nl pointer */
    int sys;		/* flag tells us never to "die" */
} VSTRING;

/* vstring implementation routines */
int vs_new(VSTRING *vs, size_t size, int sys)
{
    vs->begin= (char *)malloc(size == 0 ? 1 : size);
    if (vs->begin)
    {
	vs->end= vs->begin+size;
	vs->size= size;
	vs->p= vs->begin;
	vs->last= NULL;
	vs->sys= sys;
	return TRUE;
    }
    return FALSE;
}

void vs_destroy(VSTRING *vs)
{
    free(vs->begin);
    vs->begin= NULL;
}

int vs_realloc(VSTRING *vs, size_t size)
{
    VSTRING newvs;
    if (vs_new(&newvs,size,vs->sys))
    {
	memmove(newvs.begin,vs->begin,vs->size);
	newvs.p +=  vs->p - vs->begin;
	if (vs->last) newvs.last= newvs.begin + (vs->last - vs->begin);
	free(vs->begin);
	*vs= newvs;
	return TRUE;
    }
    /* out of memory on a realloc - not much we can do */
    return FALSE;
}

void vs_more(VSTRING *vs)
{
    /* try to realloc double - if not, realloc +256 */
    if (!vs_realloc(vs,(vs->size+1)*2) && !vs_realloc(vs,vs->size+256))
    {
	/* could not get memory - produce error message or die*/
	if (vs->sys)
	{
	    char *err= "format: vs_more: out of memory.\n";
	    vs->p= vs->begin;
	    if (vs->size > strlen(err)+1) 
	    {
		strcpy(vs->begin, err);
		vs->p+= strlen(err)+1;
	    }
	}
	else
	    die("not enough memory to hold expanding string of length %d.\n",
		vs->size+256);
    }
}


/* post increment */
char *vs_inc(VSTRING *vs)
{
    if (vs->p >= vs->end) vs_more(vs); /* moves p */
    return vs->p++;
}


/* post increment by n*/
char *vs_inc_n(VSTRING *vs, size_t n)
{
    char *ret;
    while (vs->p+n-1 >= vs->end) vs_more(vs);
    ret= vs->p;
    vs->p+= n;
    return ret;
}


/* append ordinary C string (psz) to vs */
void vs_cat(VSTRING *vs, char *psz)
{
    size_t len= strlen(psz);
    memmove(vs_inc_n(vs,len),psz,len);
}


/* prepare for returning the result - adds a NUL, then resizes it smaller
 * if it is worth the trouble */
char *vs_return(VSTRING *vs)
{
    char *p;

    /* write a NUL */
    *vs_inc(vs)= 0;

    p= vs->begin;

    /* if output is less than 80% of buffer, reallocate smaller buf */
    if ((vs->end - p)/5 > (vs->p - p)/4)
    {
	if (p= (char *)malloc(vs->p - vs->begin))
	{
	    memmove(p,vs->begin,vs->p - vs->begin);
	    vs_destroy(vs);
	}
	else p= vs->begin;
    }
    return p;
}


/* text formatting operators */

static char *pTable= NULL;

/* Character types */
#define CT_CTL 1
#define CT_BL 2
#define CT_NL 3
#define CT_LT 4
#define CT_GT 5
#define CT_AMP 6
#define CT_COLON 7
#define CT_CR 8
#define CT_TAB 9
#define CT_QUOTE 10

static void InitTable()
{
    pTable= (char *)malloc(256);
    if (!pTable) die("out of memory (256)\n");
    memset(pTable,CT_CTL,32);		/* control characters */
    memset(pTable+32,0,256-32);		/* ordinary characters */
    pTable[' ']=  CT_BL;
    pTable['\n']= CT_NL;
    pTable['<']=  CT_LT;
    pTable['>']=  CT_GT;
    pTable['&']=  CT_AMP;
    pTable[':']=  CT_COLON;
    pTable['\r']=  CT_CR;
    pTable[9]=  CT_TAB;
    pTable['"']=  CT_QUOTE;
}

/* table of matchable URL protocols */

static char * protocols[] = {
    "http",
    "https",
    "ftp",
    "gopher",
    "mailto",
    "news",
    "conf",
    "item",
    "resp",
    NULL};

#define PCOL_HTTP 0
#define PCOL_FTP 2
#define PCOL_GOPHER 3
#define PCOL MAILTO 4
#define PCOL_NEWS 5
#define PCOL_CONF 6
#define PCOL_ITEM 7
#define PCOL_RESP 8

/* prot_expand - clickify one of the above protocols.  ppin is a pointer to
 * a pointer into the input text.  It should point at the first character after
 * a ':' when called.  When returned it will point to the first character after
 * the reference.  pout is a pointer to the output string structure.   If 'wrap'
 * is non-zero, do word wrapping at that column.  Pass in a pointer to the
 * current column index and wrap delta as pcol and pwdelta if you are wrapping.
 * If 'newline' is not NULL, it points to something to output instead of
 * newlines.
 */

void prot_expand(char **ppin, VSTRING *pout,
	int wrap, int *pcol, int *pwdelta,
	char *newline)
{
    char **pp= protocols;
    char *pc;
    char *pcin;
    int protocol_len;
    while (*pp)
    {
	pc= (*pp) + (protocol_len=strlen(*pp));
	/* prepare to scan protocol backwards from colon */
	pcin= *ppin-2;
	while (--pc >= *pp && tolower(*pcin) == *pc) --pcin;
	if (pc < *pp) break;	/* protocol matched */
	++pp;		/* protocol did not match, try another */
    }
    /* the first few protocols must be followed by "//" */
    /* last 3 protocols must be followed an alphanumeric */
    if (*pp && (((*ppin)[0]=='/' && (*ppin)[1]=='/') ||
		(pp-protocols > PCOL_GOPHER)) &&
	(isalnum((*ppin)[0]) || (pp-protocols < PCOL_CONF)))
    {
	char *pint= *ppin;
        char  c1;
	Token *fmtok;

	if (pp-protocols >= PCOL_CONF) /* Backtalk-style protocol */
	{
	    char macro_name[16];
	    /* Backtalk protocols require a format macro to function */
	    /* first check to see if the format_xxxx macro exists */
	    strcpy (macro_name,"format_");
	    strcat (macro_name,*pp);
	    fmtok= get_dict(macro_name);
	    if (!fmtok || type(*fmtok) != TK_PROCEDURE) goto nonclick;
	}
	++pcin;		/* pointing to beginning of protocol name */

	/* if it is a "Backtalk" protocol, we convert it differently */
	/* Backtalk URLs are conf:, item:, and resp: */
	if (pp-protocols >= PCOL_CONF)
	{
	    /* temporary item and response numbers to pass to the filter */
	    int ni, nr;

	    char *pin1= *ppin;

	    pout->p-= strlen(*pp);

	    /* check to see if a conference name was entered */
	    if (isalpha(**ppin))
	    {
		++(*ppin);
		while (isalnum(**ppin))
		    ++(*ppin);
		pint= (char *)malloc(*ppin-pin1+1);
		memmove(pint, pin1, *ppin-pin1);
		pint[*ppin-pin1]= '\0';
		push_string(pint,FALSE);
		if (**ppin == ',' || **ppin == '.' || **ppin == ':')
		    ++(*ppin);	
	    }
	    else
	    {
		push_string(sdict(VAR_CONF), TRUE);
	    }

	    /* check if item number (ni), and response number (nr) are there */
	    nr= ni= 0;
	    while(isdigit(**ppin)) ni*=10, ni+= **ppin-'0', ++(*ppin);
	    if ((**ppin == ',' || **ppin == '.' || **ppin == ':') &&
		    isdigit((*ppin)[1]))
	    {
		++(*ppin);
		while(isdigit(**ppin))
		    nr*=10, nr+= **ppin-'0', ++(*ppin);
	    }
	    else if (pp-protocols == PCOL_RESP) nr= ni, ni= 0;

	    if (ni == 0) ni= idict(VAR_ITEM);
	    push_integer(ni);
	    push_integer(nr);
	    uniquify_token(fmtok);
	    rerun(aval(*fmtok));	/* interpret the filter macro */
	    /* <conference> <item> <response> format_xxxx <<URL> */
	    pint= pop_string();
	    vs_cat(pout,pint);
	    free(pint);							
	}
	else /* an ordinary internet protocol, expand to clickable */
	{
	    /* find end of the URL */
	    while (isgraph(**ppin) && **ppin!='"' && **ppin!='<' && **ppin!='>')
		++(*ppin);

	    while (c1= (*ppin)[-1], (ispunct(c1) && c1!='/'))
		--(*ppin);

	    if (*ppin < pint+3) {*ppin= pint; goto nonclick; }
	    pout->p-= strlen(*pp);
	    memmove(vs_inc_n(pout,9),"<a href=\"",9);		/* expand to clickable */
	    memmove(vs_inc_n(pout,*ppin-pcin),pcin,*ppin-pcin);
	    *vs_inc(pout)= '"';
	    *vs_inc(pout)= '>';
	}

	/* wrap overlong URLs */
	while (wrap && *pcol+*pwdelta+(*ppin-pcin) > wrap)
	{
	    int ii= wrap-*pcol-*pwdelta;
	    memmove(vs_inc_n(pout,ii),pcin,ii);
	    (*pcol)+= ii;
	    pcin+= ii;
	    if (newline == NULL)
    		*vs_inc(pout)= '\n';
	    else
	    {
		int l= strlen(newline);
		memmove(vs_inc_n(pout,l),newline,l);
	    }
	    (*pwdelta)= -(*pcol);
	}
	memmove(vs_inc_n(pout,*ppin-pcin),pcin,*ppin-pcin);
	memmove(vs_inc_n(pout,4),"</a>",4);
	if (pcol != NULL) (*pcol)+= *ppin-pcin;
    }
    else	/* unrecognized protocol, treat colon as ordinary */
    {
nonclick:
	*vs_inc(pout)= ':';
	if (pcol != NULL) (*pcol)++;
    }
}


/* FORMAT - does various cleanup on hunks of plain text.  if quote is TRUE, it
 * converts special HTML characters < > & to their quoted forms.  In addition
 * if expand is TRUE it will expand URLs to be clickable, and, if wrap is >0
 * it will wrap text at that many columns.  If 'sys' is false then we die on
 * error.  Otherwise not. 
 *
 * Prefix and suffix, if not NULL, are strings put at the beginning and ending
 * of the response respectively.  Each newline character is replaced with the
 * newline string, if newline is not NULL.
 *
 * Note: always frees its "input" argument when it is done.
 *       caller is responsible for freeing the result string.
 */

char *format(char *input, int quote, int expand, int wrap, int sys,
	char *prefix, char *suffix, char *newline)
{
    VSTRING out;
    int inputlen;
    int col,		/* input column number */
	wdelta,		/* wrap delta - col+wdelta = output column number */
	ct,		/* character type */
	c,		/* character itself */
	indentCol,	/* what is the column of the first graphic */
	newlinelen,	/* length of newline argument */
	prevIndentCol,	/* previous line's indentCol */
	sawGraphic,	/* seen a graphic (non-blank) on this line yet? */
	wrapping;	/* 0=not wrapping, 1 or 2=wrapping (2=hunting for graphic at NL) */
    char *pin, *lastpin; 
    
    if (!pTable) InitTable();
    if (!vs_new(&out,2*(inputlen= strlen(input)),sys))
    {
	if (sys) return "format: out of memory\n";
	else die("not enough memory to hold string of length 2*%d.\n",inputlen);
    }
    pin= input;
    wrapping= sawGraphic= indentCol= wdelta= col= 0; lastpin= NULL;

    if (prefix != NULL)
    {
	int l= strlen(prefix);
	memmove(vs_inc_n(&out,l),prefix,l);
    }
    if (newline != NULL)
    {
	newlinelen= strlen(newline);
	/* White space may be significant, so delete trailing white space */
	while (isspace(input[inputlen-1]))
	    input[--inputlen]= '\0';
    }
    
    while (*pin)
    {
	ct= pTable[c= (*pin++)&0xFF];
	if (wrapping == 2 && "xx  xxxx  "[ct]=='x')
	{
	    /* we have a graphic character while wrap logic is scanning past
	     * suppressed nl */
	    if (indentCol == prevIndentCol)
	    {
		/* continue wrapping */
		*vs_inc(&out)= ' '; /* emit blank to replace nl */
		wdelta++;
		wrapping= 1;
	    }
	    else
	    {
		/* stop wrapping because indentation changed */
		if (newline == NULL)
		    *vs_inc(&out)= '\n';
		else
		    memmove(vs_inc_n(&out,newlinelen),newline,newlinelen);
		for (wdelta= 0; wdelta < col; wdelta++) *vs_inc(&out)= ' ';
		wrapping= wdelta= 0;
	    }
	}
	if (ct)
	    switch (ct)
	    {
	    case CT_LT: 
		if (quote) vs_cat(&out,"&lt;");	/* escape the < */
		else *vs_inc(&out)= c;
		col++;
		sawGraphic= 1;
		break;
	    case CT_GT: 
		if (quote) vs_cat(&out,"&gt;");	/* escape the > */
		else *vs_inc(&out)= c;
		col++;
		sawGraphic= 1;
		break;
	    case CT_AMP: 
		if (quote) vs_cat(&out,"&amp;");	/* escape the & */
		else *vs_inc(&out)= c;
		col++;
		sawGraphic= 1;
		break;
	    case CT_QUOTE: 
		if (quote) vs_cat(&out,"&quot;");	/* escape the " */
		else *vs_inc(&out)= c;
		col++;
		sawGraphic= 1;
		break;
	    case CT_CR:			/* expand the \r char --> nothing*/
		break;
	    case CT_TAB:			/* expand the \i char */
		if (wrapping == 2) do {++col; --wdelta;} while (col%8);
		else do {*vs_inc(&out)= ' '; ++col;} while (col%8);
		if (!sawGraphic) indentCol= col;
		break;
	    case CT_CTL:			/* expand the ctl char */
		*vs_inc(&out)= '^';
		*vs_inc(&out)= (c & 0x1f) + 'A'-1;
		col+= 2;
		sawGraphic= 1;
		break;
	    case CT_NL:				/* rewind the column count */
		prevIndentCol= indentCol;
		lastpin= pin;
		out.last= out.p;
		if (wrapping)
		{
		    if (wrapping == 2)
		    {
			/* stop wrapping at paragraph boundary */
			if (newline == NULL)
			    *vs_inc(&out)= '\n';
			else
			    memmove(vs_inc_n(&out,newlinelen),
				    newline,newlinelen);
		       	wdelta= wrapping= 0;
		    }
		    else
			wrapping= 2; /* suppress NL */
		}
		if (wrapping)
		    wdelta+= col; /* don't output NL */
		else if (newline == NULL)
		    *vs_inc(&out)= '\n';
		else
		    memmove(vs_inc_n(&out,newlinelen),newline,newlinelen);
		sawGraphic= indentCol= col= 0;
		break;
	    case CT_BL:		/* remember this spot as possible wrap point */
		lastpin= pin;
		out.last= out.p;
		if (wrapping == 2)
		    --wdelta;
		else
		    *vs_inc(&out)= c;
		col++;
		if (!sawGraphic) indentCol= col;
		break;
	    case CT_COLON:		/* look for URL to make clickable */
		sawGraphic= 1;
		if (expand)
		    prot_expand(&pin, &out, wrap, &col, &wdelta, newline);
		else		/* not expanding, treat colon as ordinary */
		{
		    *vs_inc(&out)= c;
		    col++;
		};
	    }
	else /* ordinary character case */
	{
	    *vs_inc(&out) = c;
	    col++;
	    sawGraphic= 1;
	}
	
	/* now if it is beyond the limit, go back to the last wrap point
	 * and insert a nl */
	if (wrap && wdelta+col>wrap)
	{
	    if (lastpin)
	    {
		pin= lastpin;
		out.p= out.last;
	    }
	    if (newline == NULL)
		*vs_inc(&out)= '\n';
	    else
		memmove(vs_inc_n(&out,newlinelen),newline,newlinelen);
	    for (wdelta= 0; wdelta < indentCol && wdelta < wrap/2; wdelta++)
		*vs_inc(&out)= ' ';
	    wdelta= indentCol-col;
	    wrapping= 1;
	    lastpin= out.last= NULL;
	}
    }
    if (suffix)
    {
	int l= strlen(suffix);
	memmove(vs_inc_n(&out,l),suffix,l);
    }
    free(input);
    return vs_return(&out);
}


/* now we have a few HTML processing routines. First, some needed tables. */

/* list of recognized tags - must be in aphabetic order -  
 * Just because a tag is recognized doesn't mean it's permitted,
 * e.g. <TITLE> is recognized so that it can be suppressed. */

static char * TagTable[] =
{
    "A", "ABBR", "ACRONYM", "ADDRESS",
    "B", "BANNER", "BDO", "BIG", "BLOCKQUOTE", "BR",
    "CAPTION", "CENTER", "CITE", "CODE", 
    "DD", "DEL", "DIR", "DIV", "DL", "DT", 
    "EM", "EMBED", 
    "FONT", 
    "H1", "H2", "H3", "H4", "H5", "H6", 
    "I", "IMG", "INS",
    "KBD",
    "LI", 
    "MENU", 
    "NOBR", 
    "OL", 
    "P", "PRE",
    "Q",
    "S", "SAMP", "SMALL", "SPAN", "STRIKE", "STRONG", "SUB", "SUP", 
    "TABLE", "TD", "TH", "TITLE", "TR", "TT", 
    "U", "UL", 
    "WBR",
    "XMP", 
};


static size_t TagTableSize = sizeof(TagTable)/sizeof(char *);
static int DepthTable[sizeof(TagTable)/sizeof(char *)];

/* lookup_tag
 *
 * call with a address of pointer to a string to be interpreted as a TAG,
 * which may be delimited by > or whitespace. The pointer is
 * incremented to point to the delimiter.
 *
 * It returns the index. -1 if not found.
 * It is used during initialization, and also by match_tag
 */

int lookup_tag(char **pp)
{
    int min= 0;
    int max= TagTableSize-1;
    int mid;
    char *p;

    /* binary search the tag table to see if we have a match */
    while (max >= min)
    {
	char *t;

	p= *pp;
	mid= (max + min)/2;
	t= TagTable[mid];
	while (*p && toupper(*p)==*t) ++p, ++t; /* match the tag char by char */
	if (!*t && (*p == '>' || isspace(*p))) break;
	if (toupper(*p) < *t) max= mid-1; else min=mid+1;
    }

    /* we either found the tag in our table or we didn't */
    if (max < min) mid= -1;	/* we didn't */
    *pp= p;
    return mid;
}


/* These are the tags whose index needs to be known to the code. 
 * We look them up exactly once, in init_tags. Each one must be set there
 * by either a LABEL or SETLABEL macro,
 */

static int Tag_A =-1;
static int Tag_BLOCKQUOTE =-1;
static int Tag_H1 =-1;
static int Tag_IMG =-1;
static int Tag_LI =-1;
static int Tag_OL =-1;
static int Tag_P =-1;
static int Tag_TABLE =-1;
static int Tag_TD =-1;
static int Tag_TITLE =-1;
static int Tag_TR =-1;
static int Tag_UL =-1;

/* table of flags - TRUE iff the tag with this index forces line break.
 * It is initialized in init_tags, by each SET or SETLABEL macro.
 */
static char * Tags_Break_Lines =NULL;

/* similar table is TRUE for each tag that does not get closed */
static char * No_Close =NULL;

/* routine to initialize static variables needed to make sense out of TAGS */
void init_tags()
{
    if (!Tags_Break_Lines)
    {
	char *p;

#define LOOKUP(a)	lookup_tag((p=#a">",&p))
#define LABEL(a)	Tag_##a= LOOKUP(a)

	/* labeled TAGS, but not set in Tags_Break_Lines */
	LABEL(A);
	LABEL(H1);
	LABEL(IMG);

	Tags_Break_Lines= (char *)malloc(TagTableSize);
	memset (Tags_Break_Lines,0,TagTableSize);

#define SET(a)		Tags_Break_Lines[LOOKUP(a)]= TRUE;
#define SETLABEL(a)	Tags_Break_Lines[LABEL(a)]= TRUE;

	/* Tags below are all set in Tags_Break_Lines, some are labeled */
	SETLABEL(BLOCKQUOTE)
	SET(BR)
	SET(CITE)
	SET(DD)
	SET(DIV)
	SET(DL)
	SET(DT)
	SET(H1)
	SET(H2)
	SET(H3)
	SET(KBD)
	SETLABEL(LI)
	SETLABEL(OL)
	SETLABEL(P)
	SETLABEL(TABLE)
	SETLABEL(TD)
	SETLABEL(TITLE)
	SETLABEL(TR)
	SETLABEL(UL)
	SET(WBR)

	/* now init the No_Close array to TRUE for those tags that 
	 * do not have closing tags */

	No_Close= (char *)malloc(TagTableSize);
	memset (No_Close,0,TagTableSize);

#define NOCLOSE(a)	No_Close[LOOKUP(a)]= TRUE;
#define NOCLOSEL(a)	No_Close[Tag_##a]= TRUE;

	NOCLOSE(BR)
	NOCLOSE(DD)
	NOCLOSE(DT)
	NOCLOSE(EMBED)
	NOCLOSEL(IMG)
	NOCLOSEL(LI)
	NOCLOSEL(P)
	NOCLOSE(WBR)

#undef LOOKUP
#undef LABEL
#undef SET
#undef SETLABEL
#undef NOCLOSE
#undef NOCLOSEL
    }
}

/* Note: These are deliberately not recognized, and thus ignored and deleted
 * (Of course so are any other tags that are unknown to Backtalk)
 * "APPLET", "AREA", "BASEFONT", "BGSOUND", "BLINK", "BODY", "DFN", 
 * "FORM", "FRAME", "FRAMESET", "HEAD", "HR", "IMG", "INPUT", "ISINDEX", 
 * "KBD", "LINK", "MAP", "META", "NOFRAMES", "OPTION", "SAMP", 
 * "SCRIPT", "SELECT", "STYLE","TEXTAREA", "VAR", "
 * 
 * Now we are finally done initializing the TAG Labels, NoClose and LineBreaking
 * flag tables for the tags. These structures do not change once initialized.
 * 
 * Initialization of static data is done at runtime to make it easy to 
 * maintain the set of tags we handle, and their properties.
 */


/* match_tag
 *
 * search tag table to see if the tag is legit. Count it (up or down),
 * and scan to the end-of-tag marker, honoring quotes.
 *
 * it returns TRUE if a recognized tag is found and properly terminated.
 * if pTag!=NULL, it returns the number of the tag found in *pTag 
 *
 * if no tag is found (eg, no closing >) it returns FALSE and pTag is set
 * to -2.
 *
 * if a tag is found, but it is not a recognized tag, it returns false and
 * pTag is set to -1.
 *
 * *pp is advanced past the end of the tag, even if it was not recognized.
 *
 * Comments are treated as unrecognized tags, but are parsed so that *pp can
 * be advanced to the proper place.  */

int match_tag(char **pp, int *pTag)
{
    char *p= *pp;

    if (*p == '!')	/* html or sgml comment */
    {
	int sgml= p[1]=='-' && p[2]=='-';
	char *firstbrac= NULL;
	do
	{
	    while (*p && *p != '>') ++p;
	    if (*p) ++p; else break;
	    if (!firstbrac) firstbrac= p;
	} 
	while (sgml && (p[-2] != '-' || p[-3] != '-'));

	/* If we never found an "-->" to match the "<!--", stop at the
	 * first ">" we saw.  This is what mozilla does.
	 */
	if (sgml && (p[-2] != '-' || p[-3] != '-') && firstbrac)
	    p= firstbrac;

	*pp= p;
	if (pTag) *pTag= -1;
	return 0; /* treat all comments as failures to match */
    }
    else	/* a real tag - see if we recognize it */
    {
	int close= 0;
	int quote= 0;
	int tagnum;

	p= *pp;
	if (close= *p=='/') ++p;

	tagnum= lookup_tag(&p);

	/* now scan for the end of the tag */
	for (;;)
	{
	    if (*p == '\0')
	    {
		if (pTag) *pTag= -2;
		return 0;
	    }
	    else if (quote)
	    {
		if (*p == quote) quote= 0;
	    }
	    else if (*p == '\'' || *p == '"')
		quote= *p;
	    else if (*p == '>')
		break;
	    p++;
	}
	if (*p) ++p;	/* bump past the '>' */

	/* now update the counts, if it is a valid tag that needs to be
	 * counted */
	if (tagnum >= 0)
	{
	    if (
		/* disallowed <IMG> */
		(tagnum == Tag_IMG && !bdict(VAR_ALLOWGRAPHICS)) ||
		/* disallowed </TABLE> outside TABLE */
		(tagnum == Tag_TABLE && !DepthTable[Tag_TABLE] && close) ||
		/* disallowed <TR> outside TABLE */
		(tagnum == Tag_TR && !DepthTable[Tag_TABLE]) ||
		/* disallowed <TD> outside TR */
		(tagnum == Tag_TD && !DepthTable[Tag_TR]) )
	    {
		tagnum= -1;
	    }
	    else if (!No_Close[tagnum])
	    {
		if (!close) ++DepthTable[tagnum];
		else if (DepthTable[tagnum] > 0) --DepthTable[tagnum];
	    }
	    else if (close && No_Close[tagnum])
		tagnum= -1;	/* don't recognize </p> </hr> etc */
	}

	*pp= p;

	if (pTag) *pTag= tagnum;
	return tagnum >= 0;
    }
}

/* subroutine to close a tag - called below when needed */

void CloseTag (VSTRING* pOut, int tag)
{
    char *pTag= TagTable[tag];

    /* ensure closing TD before TR and TR before TABLE */
    while (tag==Tag_TR && DepthTable[Tag_TD]) CloseTag(pOut,Tag_TD);
    while (tag==Tag_TABLE && DepthTable[Tag_TR]) CloseTag(pOut,Tag_TR);

    /* close tag */
    *vs_inc(pOut)= '<';
    *vs_inc(pOut)= '/';
    while (*pTag && *pTag!='/') *vs_inc(pOut)= *pTag++;
    *vs_inc(pOut)= '>';
    --DepthTable[tag];
}

/* format_html - a subroutine, not a stack operator. Call it with a string that 
 * should contain text/html. It will sanity-check the HTML and sanitize it,
 * returning the "clean, sane" HTML in a freshly allocated string.
 * Calls free on its input, so make a copy first if you want to use input again.
 * Caller must free the output.  If 'newline' is non-null, it is a string to
 * substitute for any newline not in a tag.  If 'expand' is true, we
 * clickify URLs outside of tags.
 * Returns NULL if an internal error occurred.
 */

char *format_html(char *input, char *newline, int expand)
{
    char *pin= input;
    int newlinelen;
    char c;
    VSTRING out;

    if (newline != NULL) newlinelen= strlen(newline);

    init_tags();
    memset(DepthTable,0,sizeof(DepthTable));

    if (!vs_new(&out,strlen(input)+256,FALSE))
	 die("format_html: not enough memory for string of length %d + 256.\n",
	     strlen(input));

    while (*pin)
    {
	/* scan for tags, and check each one to see if it is permitted,
	 * and also count it */
	c= *pin++;
	if (c == '<')
	{
	    if (*pin=='/' || *pin=='!' || isalpha(*pin)) /* now we have a tag */
	    {
		char *pBeginTag= pin-1;
		int tag= -1;
		/* recognized tag, not TITLE */
		if (match_tag(&pin,&tag) && tag != Tag_TITLE )
		{
		    size_t TagSize= pin-pBeginTag;
		    memmove(vs_inc_n(&out,TagSize),pBeginTag,TagSize);
		}
		else if (tag == -2)
		{
		    /* Unclosed tag - escape the < */
		    pin= pBeginTag + 1;
		    memmove(vs_inc_n(&out,4),"&lt;",4);
		}
	    }
	    else memmove(vs_inc_n(&out,4),"&lt;",4); /* '<' not a tag, but don't trust browser, escape it */
	}
	else if (expand && c == ':' && !DepthTable[Tag_A])
	    prot_expand(&pin, &out, 0, NULL, NULL, newline);
	else if (newline != NULL && c == '\n')
	    memmove(vs_inc_n(&out,newlinelen),newline,newlinelen);
	else if (!DepthTable[Tag_TITLE])
	    *vs_inc(&out)= c;	/* copy text - suppress title text*/
    }

    free (input);

    /* we are done with sanity-checks - now sanitize:
     * clean up any unclosed tags */
    DepthTable[Tag_TITLE]= 0; /* don't need to sanitize, was suppressed */
    {
	size_t tag;

	for (tag= 0; tag < TagTableSize; tag++)
	    while (DepthTable[tag] > 0)
		CloseTag(&out,tag);
    }
    return vs_return(&out);
}

/* flush_token - a utility used by format_html2text, below */
static void flush_token(VSTRING *pToken, VSTRING *pOut, int LeftMargin, int RightMargin, int *pOutColumn, int *pNeedSep)
{
    char *p= pToken->begin;
    int count;

    while (count= pToken->p - p)
    {
	if (*pOutColumn > LeftMargin &&
		*pOutColumn + count + *pNeedSep > RightMargin)
	{
	    *vs_inc(pOut)= '\n';
	    *pOutColumn= *pNeedSep= 0;
	}

	if (*pOutColumn < LeftMargin)
	{
	    memset(vs_inc_n(pOut,LeftMargin - *pOutColumn),' ',LeftMargin - *pOutColumn);
	    *pOutColumn= LeftMargin;
	    *pNeedSep= 0;
	}

	/* emit separater if needed */
	if (*pNeedSep) *vs_inc(pOut)= ' ';
	++(*pOutColumn);

	/* emit the current token, but still wrap it at 80 if it is
	 * long enough to need it */
	if (80 - *pOutColumn < count) count= 80 - *pOutColumn;
	memmove(vs_inc_n(pOut,count),p,count);
	p+= count;
	(*pOutColumn)+= count;
	if (p >= pToken->p) break;	/* the token fit */

	*vs_inc(pOut)= '\n';	/* the token needed folding */
	*pOutColumn= 0;
    }
    pToken->p= pToken->begin;	/* empty the token VSTRING*/
    *pNeedSep= 1;
}

/* table of escape sequences for HTML characters */

static char* EscapeTable[]=
   {"LT;<","GT;>","AMP;&","NBSP; ","COPY;(C)", NULL};


/* format_html2text - a subroutine, not a stack operator. Call it with a
 * string that should contain text/html. It will produce formatted text
 * from it.
 *
 * Thus this routine functions much like a browser. It is a micro-mini-browser,
 * with the ability to understand and process only a small handful of tags.
 *
 * like the others, it frees its input when its done reading it, and produces
 * an output string which it is the caller's responsibility to free.
 */

char *format_html2text(char *input)
{
    char *pin= input;
    VSTRING out;
    VSTRING token;
    int outColumn= 0;
    int NeedSep= 0;
    int LeftMargin= 0;
    int RightMargin= 80;

    init_tags();
    memset(DepthTable,0,sizeof(DepthTable));

    if (!vs_new(&out,strlen(input),FALSE))
     die("format_html2text: not enough memory for string of length %d.\n",
	 strlen(input));
    if (!vs_new(&token,80,FALSE))
	 die("format_html2text: not enough memory for string of length 80.\n");

    while (*pin)
    {
	while (*pin && isspace(*pin)) ++pin;	/* ignore whitespace */

	/* various meta-characters - most of which we remove, some of which
	 * we fix up */
	if ((unsigned char)*pin > 127)
	{
	    switch ((unsigned char)*pin)
	    {
	    case 145:	/* Smart single left quote */
		*vs_inc(&token)= '`';
		break;

	    case 146:	/* Smart single right quote */
		*vs_inc(&token)= '\'';
		break;

	    case 147:	/* Smart double left quote */
	    case 148:	/* Smart double right quote */
		*vs_inc(&token)= '"';
		break;

	    case 169:	/* copyright */
		*vs_inc(&token)= '(';
		*vs_inc(&token)= 'c';
		*vs_inc(&token)= ')';
		break;
	    }
	    pin++;
	}

	else if (*pin=='&') /* an escape character is probable */
	{
	    char ** esc;
	    char *pe;
	    char *pi;

	    /* See if we recognize the escape sequence */
	    for (esc= EscapeTable; *esc; esc++)
	    {
		pe= *esc;
		pi= pin+1;
		while (*pe && *pe==toupper(*pi) && *pe != ';') ++pi, ++pe;
		if (*pi == ';' && *pe == ';') break;	/* we got an exact escape sequence */
	    }
	    if (*esc)	/* we got one, emit its production */
	    {
		pin= pi+1;
		for (++pe; *pe; ++pe) *vs_inc(&token)= *pe;
	    }
	    else	/* failed to match - just send out the & and move on */
	    {
		pin++;
		*vs_inc(&token)= '&';
	    }
	}
	/* see if we have a tag here */
	else if (*pin=='<')	/* a tag is probable */
	{
	    if (pin[1]=='/' || pin[1]=='!' || isalpha(pin[1])) /* now we have a tag */
	    {
		int tag;
		char *pBeginTag= pin++;

		if(match_tag(&pin,&tag)) 
		{
		    /* recognized tag - some we handle specially right now */
		    if (Tags_Break_Lines[tag])
		    {
			/* These tags must finish any earlier lines first -
			 * first flush any buffered token */
			if (token.p > token.begin)
			     flush_token(&token, &out, LeftMargin, RightMargin,
				 &outColumn, &NeedSep);


			if (outColumn)	/* end the current line if there is stuff on it */
			{
			    *vs_inc(&out)= '\n';
			    outColumn= NeedSep= 0;
			}

			if (tag==Tag_P) *vs_inc(&out)= '\n';
/*			else if (tag==Tag_HR) 
			{
			    if (LeftMargin) memset(vs_inc_n(&out,LeftMargin),' ',LeftMargin);
			    memset(vs_inc_n(&out,RightMargin-LeftMargin),'-',RightMargin-LeftMargin);
			    *vs_inc(&out)= '\n';
			}
*/			else if (tag==Tag_LI)
			{
			    if (LeftMargin>2)
				memset(vs_inc_n(&out,LeftMargin-2),' ',
					LeftMargin-2);
			    *vs_inc(&out)= "**o+-><="[(DepthTable[Tag_UL]+DepthTable[Tag_OL]) %8]; 
			    *vs_inc(&out)= ' ';
			    outColumn= LeftMargin;
			}
		    }
		    else if (tag == Tag_IMG)
		    {
			char *pi;
			int quote= 0;
			int alt= 0;

			if (token.p > token.begin)
			     flush_token(&token, &out, LeftMargin, RightMargin, &outColumn, &NeedSep);

			/* find the alt option if it is given */

			for (pi= pBeginTag; pi < pin; pi++)
			{
			    if (*pi == '"')
			    {
				if (alt && quote) break;
				quote= !quote;
			    }
			    else if (alt) *vs_inc(&token)= *pi;
			    else if (!quote && isspace(*pi) &&
				     toupper(pi[1])=='A' &&
				     toupper(pi[2])=='L' &&
				     toupper(pi[3])=='T' && pi[4]=='=')
			    {
				alt= 1;
				pi+= 4;
			    }
			}
			if (!alt) vs_cat(&token,"[IMAGE]");	/* emit default if none provided */
			if (token.p > token.begin)
			     flush_token(&token, &out, LeftMargin, RightMargin, &outColumn, &NeedSep);
		    }

		    /* recompute margins after processing a tag */
		    LeftMargin= DepthTable[Tag_BLOCKQUOTE]*7 + (DepthTable[Tag_UL]+DepthTable[Tag_OL])*4;
		    if (DepthTable[Tag_UL]+DepthTable[Tag_OL] > 0) LeftMargin+= 2;
		    RightMargin= 80 - DepthTable[Tag_BLOCKQUOTE]*7;
		    if (RightMargin < LeftMargin+15)
		    {
			RightMargin= LeftMargin+ 15;
			if (RightMargin > 80) RightMargin= 80;
		    }

		}
		else if (tag == -2)
		{
		    /* Unclosed tag - just append the < */
		    *vs_inc(&token)= '<';
		    pin= pBeginTag+1;
		}
		/* else unrecognized tag, just ignore it */
	    }
	    else
		*vs_inc(&token)= *pin++; /* '<' not a tag. append it to token */
	}

	/* otherwise just move a token into the token VSTRING */
	else while (*pin && *pin != '&' && *pin != '<' && !isspace(*pin) && (unsigned char)*pin <= 127)
	{
	    char c= *pin++;
	    *vs_inc(&token)= DepthTable[Tag_H1] ? toupper(c): c;
	}

	/* suppress text appearing between title directives */
	if (DepthTable[Tag_TITLE]) token.p= token.begin;
	/* now see if we have completed building a token, or if it needs to be
	 * forced out. */
	else if (token.p > token.begin && (isspace(*pin) || !*pin))
	     flush_token(&token, &out, LeftMargin, RightMargin, &outColumn, &NeedSep);
    }

    vs_destroy(&token);
    return vs_return(&out);
}


#ifdef CLEANUP
void free_format()
{
    if (pTable != NULL) free(pTable);
    pTable= NULL;
    if (No_Close != NULL) free(No_Close);
    No_Close= NULL;
    if (Tags_Break_Lines != NULL) free(Tags_Break_Lines);
    Tags_Break_Lines= NULL;
}
#endif


#ifndef TESTMAIN
/* these entry points are suppressed if testing */

/* FUNC_QUOTE:  <ascii-string> quote <html-string>
 * replaces reserved HTML characters with quoted versions, so that the
 * resulting text will not contain any HTML tags
 */

void func_quote()  
{ 
    push_string(format(pop_string(), TRUE, FALSE, 0, FALSE, NULL, NULL, NULL),
	    FALSE); 
}

/* FUNC_EXPAND:  <ascii-string> expand <html-string>
 * same as quote, but makes URLs clickable and wraps text at 79 cols
 */

void func_expand() 
{ 
    push_string(format(pop_string(), TRUE, TRUE , 79, FALSE, NULL, NULL, NULL),
	    FALSE); 
}

/* FUNC_WRAP:  <string> <numcols> wrap <wrapped-string>
 * Does nothing but wrap to the specified number of columns
 */

void func_wrap()
{
    int cols= pop_integer();
    push_string(format(pop_string(), FALSE, FALSE, cols, FALSE,
		NULL, NULL, NULL), FALSE);
}

/* FUNC_CLEANHTML: <html-string> clean_html <html-string>
 * Sanity check and sanitize and HTML string.
 */

void func_cleanhtml()
{
    char *clean= format_html(pop_string(),NULL,TRUE);
    if (clean == NULL) die("HTML cleanup failed");
    push_string(clean,FALSE);
}

/* FUNC_UNHTML: <html-string> clean_html <ascii-string>
 * Convert an HTML string to something plain.
 */

void func_unhtml()
{
    char *text= format_html2text(pop_string());
    if (text == NULL) die("HTML conversion failed");
    push_string(text,FALSE);
}


/* sysformat: called internally to quote a string for dumping into HTML. 
* see func_quote, above. strdups input, so caller is responsible for freeing.
* Will not call "die" or otherwise fail. Caller must still free output too. */

char *sysformat(char *input)
{
    return format(strdup(input), TRUE, FALSE, 0, TRUE, NULL, NULL, NULL);
}
#endif

/* the rest of this program is only used for testing */
#ifdef TESTMAIN
main()
{
    char *inbuf, *outbuf, line[256];
    int inlen;
    
    for (;;)
    {
	inbuf=(char *)malloc(1);
	*inbuf= '\0';
	inlen= 0;
	
	for (;;)
	{
	    if (fgets(line,256,stdin))
	    {
		int n;
		char *newinbuf;
		
		if (!strcmp(line,".\n")) break;
		n= strlen(line);
		newinbuf= (char *)malloc(inlen+n+1);
		memmove(newinbuf,inbuf,inlen);
		memmove(newinbuf+inlen,line,n);
		newinbuf[inlen+n]= '\0';
		free(inbuf);
		inbuf= newinbuf;
		inlen+= n;
	    }
	    else break;
	}
	if (!*inbuf) break;
/*	outbuf= format(inbuf,TRUE,TRUE,40,FALSE,NULL,NULL,NULL); /**/

	outbuf= format_html(inbuf,NULL,TRUE); /**/
/*	outbuf= format_html2text(inbuf); /**/
	printf("\n\n\n===============\n\n\n");
	fputs(outbuf,stdout);
	free(outbuf);
    }
    return 0;
}


#if __STDC__
void die(const char *fmt,...)
#else
void die(fmt, va_alist)
    const char *fmt;
    va_dcl
#endif
{
va_list ap;

    VA_START(ap,fmt);
    fputs("DIED:  ",stdout);
    vprintf(fmt, ap);
    fputc('\n',stdout);
    va_end(ap);
    exit(1);
}

void push_exec(int flag, char *src, ...) {}
void push_integer(int n) {}
void push_string(char *s, int copy) {}
void uniquify_token(Token *t) {};

void rerun(Array *src) {}
HashEntry sysdict[1];
char *pop_string() {return strdup("[filter results]");}
Token *get_dict(char *label)
{
    static Token t;
    t.flag= TK_PROCEDURE;
    return(&t);
}


#endif
